/* GOB C Preprocessor
 * Copyright (C) 1999 the Free Software Foundation.
 *
 * Author: George Lebl
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the  Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */
%{

#include <glib.h>

#include "y.tab.h"

static int parenth_depth = 0;
static int before_comment = INITIAL;
static int class_after_c = FALSE;
static int header_c = FALSE;

static GString *cbuf = NULL;

int line_no = 1;

static void
clear_cbuf(void)
{
	if(!cbuf) {
		cbuf = g_string_new("");
	} else {
		cbuf = g_string_assign(cbuf,"");
	}
}

static void
add_to_cbuf(char *s)
{
	if(!cbuf) {
		cbuf = g_string_new(s);
	} else {
		cbuf = g_string_append(cbuf,s);
	}
}

%}

%x COMMENT
%x C_CODE
%x C_CODE_STRING
%x CLASS_CODE
%x CLASS_CODE_I

%%

<*>\n			{ line_no++; REJECT; }

\/\/.*$			{ ; /*comment, ignore*/ }
<C_CODE>\/\/.*$		{ ; /*comment, ignore*/ }
<CLASS_CODE>\/\/.*$	{ ; /*comment, ignore*/ }
<CLASS_CODE_I>\/\/.*$	{ ; /*comment, ignore*/ }
\/\*		{BEGIN(COMMENT); before_comment = INITIAL; }
<C_CODE>\/\*	{BEGIN(COMMENT); before_comment = C_CODE; }
<CLASS_CODE>\/\*	{BEGIN(COMMENT); before_comment = CLASS_CODE; }
<CLASS_CODE_I>\/\*	{BEGIN(COMMENT); before_comment = CLASS_CODE_I; }
<COMMENT>\*\/	{BEGIN(before_comment);}
<COMMENT>.	{ ; /* comment, ignore */ }

^\%h\{		{
			BEGIN(C_CODE);
			parenth_depth = 1;
			class_after_c = FALSE;
			header_c = TRUE;
			clear_cbuf();
		}
^\%\{		{
			BEGIN(C_CODE);
			parenth_depth = 1;
			class_after_c = FALSE;
			header_c = FALSE;
			clear_cbuf();
		}
<C_CODE>^\%\}	{
			BEGIN(INITIAL);
			yylval.cbuf = cbuf;
			cbuf = NULL;
			if(header_c)
				return HCODE;
			else
				return CCODE;
		}

<C_CODE>\'\{\'		{ add_to_cbuf(yytext); }
<C_CODE>\'\\\{\'	{ add_to_cbuf(yytext); }
<C_CODE>\'\}\'		{ add_to_cbuf(yytext); }
<C_CODE>\'\\\}\'	{ add_to_cbuf(yytext); }
	
<C_CODE>\\.	{ add_to_cbuf(yytext); }
<C_CODE>\"	{
			BEGIN(C_CODE_STRING);
			add_to_cbuf(yytext);
		}
<C_CODE_STRING>\\.	{
				add_to_cbuf(yytext);
			}
<C_CODE_STRING>\"	{
				BEGIN(C_CODE);
				add_to_cbuf(yytext);
			}
<C_CODE_STRING>.	{
				add_to_cbuf(yytext);
			}

<C_CODE>\{	{
			parenth_depth++;
			add_to_cbuf(yytext);
		}
<C_CODE>\}	{
			parenth_depth--;
			if(parenth_depth<0) {
				REJECT;
			} else if(parenth_depth==0 && class_after_c) {
				BEGIN(CLASS_CODE_I);
				yylval.cbuf = cbuf;
				cbuf = NULL;
				return CCODE;
			}
			add_to_cbuf(yytext);
		}

<C_CODE>.	{ add_to_cbuf(yytext); }
<C_CODE>\n	{ add_to_cbuf(yytext); }

class		{
			BEGIN(CLASS_CODE);
			return CLASS;
		}

<CLASS_CODE>from	{return FROM;}

<CLASS_CODE_I>void	{return VOID;}
<CLASS_CODE_I>struct	{return STRUCT;}
<CLASS_CODE_I>union	{return UNION;}
<CLASS_CODE_I>enum	{return ENUM;}
<CLASS_CODE_I>signed	{return SIGNED;}
<CLASS_CODE_I>unsigned	{return UNSIGNED;}
<CLASS_CODE_I>long	{return LONG;}
<CLASS_CODE_I>short	{return SHORT;}
<CLASS_CODE_I>int	{return INT;}
<CLASS_CODE_I>float	{return FLOAT;}
<CLASS_CODE_I>double	{return DOUBLE;}
<CLASS_CODE_I>char	{return CHAR;}

<CLASS_CODE_I>public	{yylval.line = line_no; return PUBLIC;}
<CLASS_CODE_I>private	{yylval.line = line_no; return PRIVATE;}
<CLASS_CODE_I>argument	{yylval.line = line_no; return ARGUMENT;}
<CLASS_CODE_I>virtual	{yylval.line = line_no; return VIRTUAL;}
<CLASS_CODE_I>signal	{yylval.line = line_no; return SIGNAL;}
<CLASS_CODE_I>last	{return LAST;}
<CLASS_CODE_I>first	{return FIRST;}
<CLASS_CODE_I>override	{yylval.line = line_no; return OVERRIDE;}
<CLASS_CODE_I>check	{return CHECK;}
<CLASS_CODE_I>null	{return CNULL;}
<CLASS_CODE_I>type	{return TYPE;}
<CLASS_CODE_I>onerror	{return ONERROR;}
<CLASS_CODE_I>0|[1-9][0-9]*|0x[0-9a-fA-F]+|0[0-7]+|[0-9]*\.[0-9]+|\.[0-9][0-9]*	{
			yylval.id = g_strdup(yytext);
			return NUMBER;
		}
<CLASS_CODE,CLASS_CODE_I>:?[A-Za-z_][A-Za-z0-9_]*(:[A-Za-z0-9_]*)+	{
			yylval.id = g_strdup(yytext);
			return TYPETOKEN;
		}
<CLASS_CODE,CLASS_CODE_I>[A-Za-z_][A-Za-z0-9_]*	{
			yylval.id = g_strdup(yytext);
			return TOKEN;
		}

<CLASS_CODE>\{	{
			BEGIN(CLASS_CODE_I);
			return '{';
		}
<CLASS_CODE_I>\{	{
			BEGIN(C_CODE);
			parenth_depth=1;
			class_after_c = TRUE;
			return '{';
		}
<CLASS_CODE_I>\}	{
				BEGIN(INITIAL);
				return '}';
			}

<CLASS_CODE,CLASS_CODE_I,INITIAL>[\n\t ]	;  /*ignore*/

<*>.		{
			yylval.line = line_no;
			return yytext[0];
		}
